from unittest import skip
from utils import BaseTestCase


class TestWordParsing(BaseTestCase):

    def get_box(self, content):
        return next(iter(self.parse(content)))

    def get_words(self, content):
        return list(map(str, self.get_box(content).boxes))

    def test_blank(self):
        self.assertEqual(self.get_words('<p>'), [])

    def test_plain_words(self):
        words = self.get_words('<p>hello world, bericht!')
        self.assertEqual(words, ['hello', 'world,', 'bericht!'])

    @skip
    def test_styled_words(self):
        words = self.get_words('<p>hello <i>world<b>,</b></i> <b>bericht</b><i>!</i>')
        self.assertEqual(words, ['hello', 'world,', 'bericht!'])


class TestWordWrapping(BaseTestCase):

    def get_box(self, content):
        return next(iter(self.parse(content)))

    def get_lines(self, box):
        return [list(map(str, line.words)) for line in box.lines]

    def test_blank(self):
        box = self.get_box('<p>')
        box.wrap(None, 200)
        self.assertEqual(box.lines, [])

    @skip
    def test_wrap_styled_words(self):
        box = self.get_box('<p>hello <i>world<b>,</b></i> <b>bericht</b><i>!</i>')
        box.wrap(None, 125)
        self.assertEqual(self.get_lines(box), [['hello', 'world,', 'bericht!']])
        box.wrap(None, 75)
        self.assertEqual(self.get_lines(box), [['hello', 'world,'], ['bericht!']])
        box.wrap(None, 50)
        self.assertEqual(self.get_lines(box), [['hello'], ['world,'], ['bericht!']])



from unittest import TestCase
from bericht.html import HTMLParser, CSS


def parse_html(html, css=None):
    return HTMLParser(
        lambda: iter(html if isinstance(html, list) else [html]),
        css if isinstance(css, CSS) else CSS(css or '')
    )


class BaseTestCase(TestCase):

    @staticmethod
    def parse(html, css=None):
        return parse_html(html, css)



from unittest import TestCase
from bericht.style import *

S = Style.default()


class TestCache(TestCase):

    def test_caching(self):
        bolded1 = S.set(bold=True)
        bolded2 = S.set(bold=True)
        self.assertEqual(bolded1, bolded2)
        self.assertIs(bolded1, bolded2)


class TestValidation(TestCase):

    def test_vertical_align(self):
        self.assertEqual(S.vertical_align, VerticalAlign.top)
        vert = S.set(vertical_align=VerticalAlign.bottom)
        self.assertEqual(vert.vertical_align, VerticalAlign.bottom)
        with self.assertRaises(AssertionError):
            S.set(vertical_align="foo")

    def test_page_break_inside(self):
        self.assertEqual(S.page_break_inside, True)
        pb = S.set(page_break_inside=False)
        self.assertEqual(pb.page_break_inside, False)
        with self.assertRaises(AssertionError):
            S.set(page_break_inside="foo")



from unittest import TestCase
from reportlab.pdfgen.canvas import Canvas

from bericht.style import *
from bericht.table import *
from bericht.text import *


S = Style.default()
hello_width = Word(S, 'hello 99').width
hello_height = S.leading


def hello(start, end=None, row=None, col=None):
    hellos = ' '.join('hello {}'.format(i) for i in (range(start, end) if end else range(start)))
    return 'Row {} Col {}: {}'.format(row, col, hellos) if row and col else hellos


def hello_p(start, end=None, row=None, col=None):
    return para(hello(start, end, row, col))


def row_lines(row, width, height):
    row.wrap(width, height)
    cells = []
    for cell in row.content:
        if isinstance(cell, Span):
            cells.append(cell)
            continue
        lines = []
        for p in cell.content:
            if isinstance(p, Paragraph):
                for line in p.content:
                    lines.append(' '.join(str(w) for w in line))
            elif isinstance(p, Text):
                lines.append(p.text)
            else:
                raise ValueError
        cells.append(lines)
    return cells


class RecordingParagraph(Paragraph):

    def __init__(self, words, style):
        super().__init__(words, style)
        self.draw_call = None

    def drawOn(self, canvas, x, y, _sW=None):
        self.draw_call = (x, y)
        super().drawOn(canvas, x, y, _sW)


def rpara(text, style):
    return RecordingParagraph.from_string(text, style)


class TestCell(TestCase):

    def setUp(self):
        self.text = 'hello'
        self.p = rpara(self.text, S)
        self.width = self.p.width
        self.height = S.leading

    def test_frame(self):
        frame = S.set(
            margin_top=1,
            margin_right=2,
            margin_bottom=3,
            margin_left=4,
            border_top_width=5,
            border_right_width=6,
            border_bottom_width=7,
            border_left_width=8,
            padding_top=9,
            padding_right=10,
            padding_bottom=11,
            padding_left=12,
        )
        cell = Cell([], frame)
        cell.collapsed = False
        self.assertEqual(cell.frame_top, 14)
        self.assertEqual(cell.frame_right, 16)
        self.assertEqual(cell.frame_bottom, 18)
        self.assertEqual(cell.frame_left, 20)
        self.assertEqual(cell.frame_width, 36)
        self.assertEqual(cell.frame_height, 32)
        cell.collapsed = True
        self.assertEqual(cell.frame_top, 11.5)
        self.assertEqual(cell.frame_right, 13)
        self.assertEqual(cell.frame_bottom, 14.5)
        self.assertEqual(cell.frame_left, 16)
        self.assertEqual(cell.frame_width, 29)
        self.assertEqual(cell.frame_height, 26)

    def test_wrap(self):
        cell = Cell([self.p], S.set(padding_top=3, padding_bottom=3))
        self.assertEqual(cell.wrap(50), (50, self.height + 6))
        self.assertEqual(cell.width, 50)
        self.assertEqual(cell.height, self.height + 6)
        self.assertEqual(cell.content_heights, [self.height])

    def test_wrap_multiple_flowables(self):
        cell = Cell(
            [hello_p(10), hello_p(20), hello_p(10)],
            S.set(padding_top=3, padding_bottom=3))
        self.assertEqual(cell.wrap(50), (50, 560 + 6))
        self.assertEqual(cell.width, 50)
        self.assertEqual(cell.height, 560 + 6)
        self.assertEqual(cell.content_heights, [140, 280, 140])

    def test_no_split(self):
        cell = Cell([hello_p(10)], S)
        cells = cell.split(50, 140)
        self.assertEqual(len(cells), 1)
        self.assertEqual(cells[0], cell)
        self.assertEqual(cells[0].was_split, False)

    def test_split(self):
        cell = Cell([hello_p(10)], S)
        cells = cell.split(50, 70)
        self.assertEqual(len(cells), 2)
        self.assertEqual(cells[0].was_split, True)
        self.assertEqual(cells[1].was_split, True)
        top, bottom = [str(cell.content[0]) for cell in cells]
        self.assertEqual(top, "hello 0 hello 1 hello 2 hello 3 hello 4")
        self.assertEqual(bottom, "hello 5 hello 6 hello 7 hello 8 hello 9")

    def test_draw(self):
        canvas = Canvas(None)
        p1 = rpara(self.text, S)
        p2 = rpara(self.text, S)
        cell = Cell([p1, p2], S.set(padding_left=1, padding_top=3, padding_bottom=3))
        # natural height, aligned top
        cell.wrap(30)
        cell.drawOn(canvas, 0, 0)
        self.assertEqual(p1.draw_call, (1, 17))
        self.assertEqual(p2.draw_call, (1, 3))
        # force different height
        cell.height = 100
        cell.drawOn(canvas, 0, 0)
        self.assertEqual(p1.draw_call, (1, 83))
        self.assertEqual(p2.draw_call, (1, 69))
        # height 100 and aligned to bottom
        cell.style = cell.style.set(vertical_align=VerticalAlign.bottom)
        cell.drawOn(canvas, 0, 0)
        self.assertEqual(p1.draw_call, (1, 17))
        self.assertEqual(p2.draw_call, (1, 3))
        # middle aligned
        cell.style = cell.style.set(vertical_align=VerticalAlign.middle)
        cell.drawOn(canvas, 0, 0)
        self.assertEqual(p1.draw_call, (1, 50))
        self.assertEqual(p2.draw_call, (1, 36))


class TestRow(TestCase):

    def test_frame(self):
        # rows only care about border_top and
        # border_bottom when in collapsed mode,
        # everything else is ignored
        frame = S.set(
            margin_top=1,
            margin_right=2,
            margin_bottom=3,
            margin_left=4,
            border_top_width=5,
            border_right_width=6,
            border_bottom_width=7,
            border_left_width=8,
            padding_top=9,
            padding_right=10,
            padding_bottom=11,
            padding_left=12,
        )
        row = Row([], frame)
        row.collapsed = False
        self.assertEqual(row.frame_top, 0)
        self.assertEqual(row.frame_right, 0)
        self.assertEqual(row.frame_bottom, 0)
        self.assertEqual(row.frame_left, 0)
        self.assertEqual(row.frame_width, 0)
        self.assertEqual(row.frame_height, 0)
        row.collapsed = True
        self.assertEqual(row.frame_top, 5)
        self.assertEqual(row.frame_right, 0)
        self.assertEqual(row.frame_bottom, 7)
        self.assertEqual(row.frame_left, 0)
        self.assertEqual(row.frame_width, 0)
        self.assertEqual(row.frame_height, 12)

    def test_wrap(self):
        row = Row(
            [Cell([hello_p(10)], S)],
            S.set(border_top_width=3, border_bottom_width=3)
        )
        row.collapsed = True
        self.assertEqual(row.wrap([50]), (50, 140 + 6))
        self.assertEqual(row.width, 50)
        self.assertEqual(row.height, 140 + 6)
        self.assertEqual(row.content_widths, [50])
        self.assertEqual(row.column_widths, [50])

    def test_wrap_equalized_height(self):
        row = Row([
            Cell([hello_p(10)], S),
            Cell([hello_p(20)], S),
            Cell([hello_p(10)], S)
        ], S)
        self.assertEqual(row.wrap([50, 50, 50]), (150, 280))
        self.assertEqual(row.width, 150)
        self.assertEqual(row.height, 280)
        # override individual cell heights to match the tallest cell
        self.assertEqual(row.content[0].height, 280)
        self.assertEqual(row.content[1].height, 280)
        self.assertEqual(row.content[2].height, 280)
        self.assertEqual(row.content_widths, [50, 50, 50])
        self.assertEqual(row.column_widths, [50, 50, 50])

    def test_wrap_with_span(self):
        row = Row([
            Cell([hello_p(10)], S),
            Span.col,
            Cell([hello_p(10)], S)
        ], S)
        self.assertEqual(row.wrap([50, 50, 50]), (150, 140))
        self.assertEqual(row.width, 150)
        self.assertEqual(row.height, 140)
        self.assertEqual(row.content[0].height, 140)
        self.assertEqual(row.content[1].height, 140)
        self.assertEqual(row.content_widths, [100, 50])
        self.assertEqual(row.column_widths, [50, 50, 50])

    def test_no_split(self):
        row = Row([
            Cell([hello_p(10)], S),
            Cell([hello_p(10, 20)], S)
        ], S)
        rows = row.split([50, 50], 140)
        self.assertEqual(len(rows), 1)
        self.assertEqual(rows[0], row)

    def test_split(self):
        row = Row([
            Cell([hello_p(10)], S),
            Cell([hello_p(10, 20)], S)
        ], S)
        width, height = [hello_width * 5, hello_width * 6], hello_height
        rows = row.split(width, height)
        self.assertEqual(len(rows), 2)
        self.assertEqual(rows[0].was_split, True)
        self.assertEqual(rows[1].was_split, True)
        top1, top2 = row_lines(rows[0], width, height)
        bottom1, bottom2 = row_lines(rows[1], width, height)
        self.assertEqual(top1, ["hello 0 hello 1 hello 2 hello 3 hello 4"])
        self.assertEqual(bottom1, ["hello 5 hello 6 hello 7 hello 8 hello 9"])
        self.assertEqual(top2, ["hello 10 hello 11 hello 12 hello 13 hello 14"])
        self.assertEqual(bottom2, ["hello 15 hello 16 hello 17 hello 18 hello 19"])

    def test_split_w_span(self):
        row = Row([
            Cell([hello_p(10)], S),
            Span.col,
            Cell([hello_p(10, 20)], S)
        ], S)
        width, height = [hello_width * 3, hello_width * 2, hello_width * 6], hello_height
        rows = row.split(width, height)
        self.assertEqual(len(rows), 2)
        top1, top2 = row_lines(rows[0], width, height)
        bottom1, bottom2 = row_lines(rows[1], width, height)
        self.assertEqual(top1, ["hello 0 hello 1 hello 2 hello 3 hello 4"])
        self.assertEqual(bottom1, ["hello 5 hello 6 hello 7 hello 8 hello 9"])
        self.assertEqual(top2, ["hello 10 hello 11 hello 12 hello 13 hello 14"])
        self.assertEqual(bottom2, ["hello 15 hello 16 hello 17 hello 18 hello 19"])

    def test_split_w_span_should_fail_for_lack_of_vertical_space(self):
        """
        Ask row to split but don't give enough height to fit a single
        line of text.
        Tests a regression where an empty row was considered not-empty
        because Span.col evaluated to not None.
        """
        row = Row([Cell([hello_p(10)], S), Span.col], S)
        rows = row.split([hello_width * 3, hello_width * 2], hello_height * .9)
        self.assertEqual(len(rows), 0)

    def test_split_with_bottom_alignment(self):
        """
            top   |         |  bottom
        -------------------------------
        | hello 0 | hello 0 |         |
        | hello 1 | hello 1 | hello 0 |
        | hello 2 | hello 2 | hello 1 |  <-- split above row
        |         | hello 3 | hello 2 |
        -------------------------------
        """
        row = Row([
            Cell([hello_p(3)], S),
            Cell([hello_p(4)], S),
            Cell([hello_p(3)], S.set(vertical_align=VerticalAlign.bottom))
        ], S)
        width, height = [hello_width+1]*3, 2 * hello_height + 1
        rows = row.split(width, height)
        self.assertEqual(len(rows), 2)
        top1, top2, top3 = row_lines(rows[0], width, height)
        self.assertEqual(top1, ["hello 0", "hello 1"])
        self.assertEqual(top2, ["hello 0", "hello 1"])
        self.assertEqual(top3, ["hello 0"])
        bottom1, bottom2, bottom3 = row_lines(rows[1], width, height)
        self.assertEqual(bottom1, ["hello 2"])
        self.assertEqual(bottom2, ["hello 2", "hello 3"])
        self.assertEqual(bottom3, ["hello 1", "hello 2"])


class TestTable(TestCase):

    def test_wrap_single_cell(self):
        table = Table([1], [0], [Row([Cell([hello_p(10)], S)], S)], S)
        self.assertEqual(table.wrap(50, 10), (50, 142))
        self.assertEqual(table.content_heights, [142])

    def test_wrap_uses_tallest_column_in_each_row(self):
        table = Table([1, 1], [0, 0], [
            Row([Cell([hello_p(10)], S), Cell([hello_p(20)], S)], S),
            Row([Cell([hello_p(20)], S), Cell([hello_p(10)], S)], S),
        ], S)
        self.assertEqual(table.wrap(100, 10), (100, 564))
        self.assertEqual(table.content_heights, [282, 282])

    def test_no_split_single_cell(self):
        table = Table([1], [0], [Row([Cell([hello_p(10)], S)], S)], S)
        tables = table.split(50, 142)
        self.assertEqual(len(tables), 1)
        self.assertEqual(tables[0], table)

    def test_split_single_cell(self):
        table = Table([1], [0], [Row([Cell([hello_p(10)], S)], S)], S)
        width, height = 5 * hello_width, hello_height + 2
        tables = table.split(width, height)
        self.assertEqual(len(tables), 2)
        top, bottom = [row_lines(table.content[0], [width], height)[0] for table in tables]
        self.assertEqual(top, ["hello 0 hello 1 hello 2 hello 3 hello 4"])
        self.assertEqual(bottom, ["hello 5 hello 6 hello 7 hello 8 hello 9"])

    def test_clean_split_between_rows(self):
        rows = [
            Row([Cell([hello_p(10)], S)], S),
            Row([Cell([hello_p(10)], S)], S)
        ]
        table = Table([1], [0], rows, S)
        tables = table.split(50, 142)
        self.assertEqual(len(tables), 2)
        self.assertEqual(tables[0].content, [rows[0]])
        self.assertEqual(tables[1].content, [rows[1]])

    def test_calculate_ratios(self):
        table = Table([0, 1, 0], [25, 0, 100], [], S)
        self.assertEqual(table.content_widths, None)
        table.calculate_ratio_columns(200)
        self.assertEqual(table.content_widths, [25, 75, 100])

    def test_header_footer_split(self):
        thead = [  # height: 16
            Row([Cell([static('heading', S)], S)], S)
        ]
        tbody = [  # height: 142
            Row([Cell([hello_p(10)], S)], S),
            Row([Cell([hello_p(10)], S)], S)
        ]
        tfoot = [  # height: 16
            Row([Cell([static('footer', S)], S)], S)
        ]
        table = Table([1], [0], tbody, S, thead, tfoot)
        tables = table.split(50, 16 + 142 + 16)
        self.assertEqual(len(tables), 2)
        self.assertEqual(list(tables[0].rows), [thead[0], tbody[0], tfoot[0]])
        self.assertEqual(list(tables[1].rows), [thead[0], tbody[1], tfoot[0]])


class TestBuilder(TestCase):

    def test_basic(self):
        builder = TableBuilder([1, 0, 0], S)
        builder.row(para('hello'), 'world', 99)
        tbl = builder.table
        tbl.wrap(100)
        rows = tbl.content
        self.assertEqual(len(rows), 1)
        cells = rows[0].content
        self.assertEqual(len(cells), 3)
        self.assertIsInstance(cells[0].content[0], Paragraph)
        self.assertIsInstance(cells[1].content[0], Text)
        self.assertIsInstance(cells[2].content[0], Text)
        self.assertEqual(row_lines(tbl.content[0], tbl.content_widths, 100), [
            ['hello'], ['world'], ['99']
        ])

    def test_span(self):
        builder = TableBuilder([1, 1, 1], S)
        builder.row('hello', Span.col, 'world')
        tbl = builder.table
        tbl.wrap(150)
        rows = tbl.content
        self.assertEqual(len(rows), 1)
        row = rows[0]
        self.assertEqual(len(row.columns), 3)
        self.assertEqual(len(row.content), 2)
        self.assertEqual(row.content_widths, [97, 48])
        self.assertEqual(row_lines(row, tbl.content_widths, 100), [
            ['hello'], ['world']
        ])

    def test_bigger_table_with_spanning(self):
        builder = TableBuilder([0, 1, 0, 0], S)
        builder.row('col 1', 'col 2', 'col 3', 'col 4')
        builder.row('row 1', hello_p(10), Span.col, 99)
        builder.row('row 1.1', 'stuff', 'more stuff', 11)
        tbl = builder.table

        tbl.wrap(500)
        rows = tbl.content
        self.assertEqual(len(rows), 3)
        row1, row2, row3 = rows
        self.assertEqual(len(row1.content), 4)
        self.assertEqual(len(row2.content), 3)
        self.assertEqual(len(row3.content), 4)
        self.assertEqual(row_lines(row1, tbl.content_widths, 1000), [
            ['col 1'], ['col 2'], ['col 3'], ['col 4']
        ])
        self.assertEqual(row_lines(row2, tbl.content_widths, 1000), [
            ['row 1'], [hello(10)], ['99']
        ])
        self.assertEqual(row_lines(row3, tbl.content_widths, 1000), [
            ['row 1.1'], ['stuff'], ['more stuff'], ['11']
        ])

        # with table split
        tbl.wrap(305)
        tbl1, tbl2 = tbl.split(305, 40)
        tbl1.wrap(305)
        row1, row2 = tbl1.content
        self.assertEqual(len(row1.content), 4)
        self.assertEqual(len(row2.content), 3)
        self.assertEqual(row_lines(row1, tbl.content_widths, 1000), [
            ['col 1'], ['col 2'], ['col 3'], ['col 4']
        ])
        self.assertEqual(row_lines(row2, tbl.content_widths, 1000), [
            ['row 1'], [hello(6)], ['99']
        ])
        tbl2.wrap(300)
        row1, row2 = tbl2.content
        self.assertEqual(len(row1.content), 3)
        self.assertEqual(len(row2.content), 4)
        self.assertEqual(row_lines(row1, tbl.content_widths, 1000), [
            [], [hello(6, 10)], []
        ])
        self.assertEqual(row_lines(row2, tbl.content_widths, 1000), [
            ['row 1.1'], ['stuff'], ['more stuff'], ['11']
        ])



from bericht.html import HTMLParser
from bericht.pdf import PDFStreamer


def html_gen():
    yield '<p>begin hello world<br>'
    for i in range(1, 3):
        yield '{} hello world<br>'.format(i)
    yield 'end hello world</p>'
    yield '<p>new paragraph</p>'
    yield """<table>
  <colgroup>
    <col width="1*" />
    <col width="0*">April 6, 2017</col>
  </colgroup>
  <tr>
    <td><p>Angebot (2.1)</p></td>
    <td><p>April 6, 2017</p></td>
  </tr>
</table>"""
    yield '<p>now for big table</p>'
    yield '<p>now for big table</p>'
    yield '<p>now for big table</p>'
    yield '<table>'
    yield '<colgroup>'
    yield '<col width="0*">row 99 col 99</col>'
    yield '<col width="3*">'
    yield '<col width="1*">'
    yield '</colgroup>'
    yield '<thead>'
    yield '<tr><td><p>name</td><td><p>description</td><td><p>row</td></tr>'
    yield '<tr><td></td><td colspan="2"><p>multi row header</td></tr>'
    yield '</thead>'
    #yield '<tfoot>'
    #yield '<tr><td colspan="3"><p>the footer</td></tr>'
    #yield '</tfoot>'
    for i in range(1, 30):
        yield '<tr><td><p>row {} col 1</p></td><td><p>row {} col 2 {}</p></td><td><p>{}</p></td></tr>'.format(i, i, 'blah blah' * 11, i)
    yield '</table>'
    yield '<p>last p</p>'


def test():
    with open('test.pdf', 'wb') as pdf:
        for chunk in PDFStreamer(HTMLParser(html_gen())):
            print(chunk)
            pdf.write(chunk)

test()



from os.path import dirname, join
from tinycss2 import (
    parse_stylesheet, parse_declaration_list,
    ast, nth, color3
)

from bericht.html.style import Style, BorderCollapse, BorderStyle
from bericht.html.box import Box

__all__ = ('CSS',)

inch = 72.0
cm = inch / 2.54
mm = cm * 0.1

units = {
    'px': 1,
    'pt': 1,
    'mm': mm,
    'cm': cm,
    'inch': inch
}


def convert(value):
    if isinstance(value, ast.NumberToken):
        return value.value
    if value.unit in units:
        return value.value * units[value.unit]
    raise NotImplementedError(
        'Dimension unit "{}" is not implemented yet.'.format(value.unit)
    )


class Combinator:

    DESCENDANT = 1          # E F    an F element descendant of an E element
    CHILD = 2               # E > F  an F element child of an E element
    NEXT_SIBLING = 3        # E + F  an F element immediately preceded by an E element
    SUBSEQUENT_SIBLING = 4  # E ~ F  an F element preceded by an E element

    __slots__ = ('type', 'simple_selectors')

    def __init__(self):
        self.type = None
        self.simple_selectors = []

    @staticmethod
    def select_tag(tag):
        return lambda node: tag == node.tag

    @staticmethod
    def select_id(node_id):
        return lambda node: node_id == node.id

    @staticmethod
    def select_class(name):
        return lambda node: name in node.classes

    @staticmethod
    def select_attributes(attrs):
        assert all(isinstance(attr, ast.IdentToken) for attr in attrs)

        def attr_test(node):
            for attr in attrs:
                if attr.lower_value not in node.attrs:
                    return False
            return True

        return attr_test

    @staticmethod
    def select_position(a_b):
        a, b = a_b

        def position_test(node):
            if a == 0:  # nth-child(0n+b)
                return node.position == b
            else:  # nth-child(an+b)
                return node.position >= b and (node.position-b) % a == 0

        return position_test

    @staticmethod
    def select_last_child(last_child):
        def is_last_child(node):
            return last_child == node.last
        return is_last_child

    def add(self, selector, value, negate=False):
        matcher = selector(value)
        if negate:
            self.simple_selectors.append(lambda node: not matcher(node))
        else:
            self.simple_selectors.append(matcher)

    def matches(self, node):
        for simple_selector in self.simple_selectors:
            if not simple_selector(node):
                return False
        return True


class Selector:

    __slots__ = ('combinators',)

    def __init__(self, combinators):
        self.combinators = combinators

    def matches(self, node):
        for combinator in reversed(self.combinators):
            if not combinator.matches(node):
                return False
            node = node.parent
        return True


class Declarations:

    def __init__(self, content):
        self.content = content
        self.attrs = {}
        for declaration in parse_declaration_list(content, skip_comments=True, skip_whitespace=True):
            if isinstance(declaration, ast.AtRule):
                if declaration.at_keyword == 'bottom-right':
                    at_dec = Declarations(declaration.content)
                    self.attrs['page_bottom_right_content'] = at_dec.attrs['content']
                else:
                    raise NotImplementedError
            elif declaration.name == 'content':
                self.attrs[declaration.name.replace('-', '_')] = parse_content_value(declaration.value)
            else:
                attr = declaration.name.replace('-', '_')
                value_parts = parse_value_parts(declaration.value)
                if declaration.name in ('color', 'background-color'):
                    self.attrs[attr] = color3.parse_color(value_parts[0])
                elif declaration.name in ('margin', 'padding'):
                    apply_box_values(
                        self.attrs, declaration.name+'_{}',
                        list(map(convert, value_parts))
                    )
                elif declaration.name == 'border-color':
                    apply_box_values(self.attrs, 'border_{}_color', list(map(color3.parse_color, value_parts)))
                elif declaration.name == 'border-width':
                    apply_box_values(self.attrs, 'border_{}_width', list(map(convert, value_parts)))
                elif declaration.name == 'border-collapse':
                    self.attrs[attr] = getattr(BorderCollapse, value_parts[0].value)
                elif declaration.name == 'border-style':
                    apply_box_values(self.attrs, 'border_{}_style',
                                     [getattr(BorderStyle, v.value) for v in value_parts])
                elif declaration.name == 'border-spacing':
                    if len(value_parts) == 1:
                        self.attrs['border_spacing_horizontal'] = value_parts[0].value
                        self.attrs['border_spacing_vertical'] = value_parts[0].value
                    else:
                        (self.attrs['border_spacing_horizontal'],
                         self.attrs['border_spacing_vertical']) = value_parts
                elif declaration.name.startswith('border-'):
                    if declaration.name.endswith('-style'):
                        self.attrs[attr] = getattr(BorderStyle, value_parts[0].value)
                    elif declaration.name.endswith('-color'):
                        self.attrs[attr] = color3.parse_color(value_parts[0])
                    elif declaration.name.endswith('-width'):
                        self.attrs[attr] = convert(value_parts[0])
                elif isinstance(value_parts[0], ast.DimensionToken):
                    self.attrs[attr] = convert(value_parts[0])
                else:
                    self.attrs[attr] = value_parts[0].value

    def apply(self, node):
        if node.style:
            node.style = node.style.set(**self.attrs)
        else:
            parent = Style.default() if node.parent is None else node.parent.style
            node.style = parent.set(**self.attrs)


def apply_box_values(attrs, attr, values):
    if len(values) == 1:
        attrs[attr.format('top')] = values[0]
        attrs[attr.format('right')] = values[0]
        attrs[attr.format('bottom')] = values[0]
        attrs[attr.format('left')] = values[0]
    elif len(values) == 2:
        attrs[attr.format('top')] = values[0]
        attrs[attr.format('right')] = values[1]
        attrs[attr.format('bottom')] = values[0]
        attrs[attr.format('left')] = values[1]
    elif len(values) == 3:
        attrs[attr.format('top')] = values[0]
        attrs[attr.format('right')] = values[1]
        attrs[attr.format('bottom')] = values[2]
        attrs[attr.format('left')] = values[1]
    else:
        (attrs[attr.format('top')],
         attrs[attr.format('right')],
         attrs[attr.format('bottom')],
         attrs[attr.format('left')]) = values


def parse_value_parts(tokens):
    return list(filter(lambda token: token.type not in ('whitespace', 'comment'), tokens))


def parse_content_value(value):
    parts = []
    for part in value:
        if isinstance(part, ast.StringToken):
            parts.append(part.value)
        elif isinstance(part, ast.FunctionBlock):
            if part.lower_name == 'counter' and part.arguments[0].lower_value == 'page':
                parts.append(lambda page: page.page_number)
            else:
                raise NotImplementedError
        elif isinstance(part, ast.WhitespaceToken):
            pass
        else:
            raise NotImplementedError

    return lambda page: ''.join(
        [p if isinstance(p, str) else str(p(page)) for p in parts]
    )


def parse_selectors(prelude):
    selectors = []
    combinator = None
    combinators = []
    prefix = None

    for token in prelude:
        if isinstance(token, ast.WhitespaceToken):
            combinator = None
        elif token == ',':
            selectors.append(Selector(combinators))
            combinator = None
            combinators = []
        elif token == '>':
            combinators[-1].type = Combinator.CHILD
            combinator = None
        elif token == '+':
            combinators[-1].type = Combinator.NEXT_SIBLING
            combinator = None
        elif token == '~':
            combinators[-1].type = Combinator.SUBSEQUENT_SIBLING
            combinator = None
        else:
            if not combinator:
                combinator = Combinator()
                combinators.append(combinator)
            if token in ('.', ':'):
                if not prefix:
                    prefix = token.value
                else:
                    prefix += token.value
            elif isinstance(token, ast.IdentToken):
                if prefix is None:
                    combinator.add(combinator.select_tag, token.value)
                elif prefix == '.':
                    combinator.add(combinator.select_class, token.value)
                elif prefix == ':':
                    if token.value == 'first':
                        combinator.add(combinator.select_position, (0, 1))
                    elif token.value in ('last', 'last-child'):
                        combinator.add(combinator.select_last_child, True)
                    else:
                        raise NotImplementedError
                else:
                    raise NotImplementedError
                prefix = None
            elif isinstance(token, ast.HashToken):
                prefix = None
                combinator.add(combinator.select_id, token.value)
            elif isinstance(token, ast.FunctionBlock):
                prefix = None
                if token.name == 'nth-child':
                    combinator.add(combinator.select_position, nth.parse_nth(token.arguments))
                elif token.name == 'nth-of-type':
                    raise NotImplementedError
            elif isinstance(token, ast.SquareBracketsBlock):
                prefix = None
                combinator.add(combinator.select_attributes, token.content)
            else:
                raise NotImplementedError

    selectors.append(Selector(combinators))

    return selectors


class CSS:

    def __init__(self, src):
        self.src = src
        self.rules = []
        with open(join(dirname(__file__), 'html.css'), 'r') as html_css:
            self.parse(html_css.read())
        self.parse(src)

    def parse(self, src):
        for rule in parse_stylesheet(src, skip_comments=True, skip_whitespace=True):
            if isinstance(rule, ast.AtRule) and rule.at_keyword == 'page':
                selectors = parse_selectors(rule.prelude)
                for selector in selectors:
                    for combinator in selector.combinators:
                        combinator.add(combinator.select_tag, '@page')
                    if not selector.combinators:
                        combinator = Combinator()
                        combinator.add(combinator.select_tag, '@page')
                        selector.combinators.append(combinator)
                self.rules.append(
                    (selectors, Declarations(rule.content))
                )
            elif isinstance(rule, ast.QualifiedRule):
                self.rules.append(
                    (parse_selectors(rule.prelude), Declarations(rule.content))
                )

    def apply(self, node):
        for selectors, declarations in self.rules:
            for selector in selectors:
                if selector.matches(node):
                    declarations.apply(node)

    def apply_recursively(self, node):
        self.apply(node)
        for child in node.children:
            if isinstance(child, Box):
                self.apply_recursively(child)



from unittest import TestCase

from reportlab.pdfgen.canvas import Canvas

from bericht.html import HTMLParser
from bericht.node import Paragraph, Break


class BaseTestCase(TestCase):

    def draw(self, p):
        canvas = Canvas(None)
        p.wrapOn(canvas, 200, 100)
        p.canv = canvas
        p.draw()
        return canvas._code[0]

    def assertPDF(self, input, output):
        self.assertEqual(self.draw(input), ' '.join(output))

    @staticmethod
    def parse(html):
        return next(iter(HTMLParser(iter(html if isinstance(html, list) else [html]))))


class TestFormattingTagParsing(BaseTestCase):

    def test_bold_italic_underline(self):

        p = self.parse(
            '<p><b>hello</b> <i>world</i>, <u>bericht</u>! '
            '<strong>strong</strong> <b><i><u>biu</u></i></b></p>'
        )

        self.assertIsInstance(p, Paragraph)

        words = p.words
        self.assertEqual(len(words), 5)

        bold, italic, underline, strong, biu = (w.parts[0] for w in words)

        self.assertEqual(bold.style.bold, True)
        self.assertEqual(bold.style.italic, False)
        self.assertEqual(bold.style.underline, False)

        self.assertEqual(strong.style.bold, True)
        self.assertEqual(strong.style.italic, False)
        self.assertEqual(strong.style.underline, False)

        self.assertEqual(italic.style.bold, False)
        self.assertEqual(italic.style.italic, True)
        self.assertEqual(italic.style.underline, False)

        self.assertEqual(underline.style.bold, False)
        self.assertEqual(underline.style.italic, False)
        self.assertEqual(underline.style.underline, True)

        self.assertEqual(biu.style.bold, True)
        self.assertEqual(biu.style.italic, True)
        self.assertEqual(biu.style.underline, True)

    def test_chunking(self):
        p = self.parse(["<p>", "I", "’ve ", "I’ve</p>"])
        words = p.words
        self.assertEqual(len(words), 2)
        self.assertEqual(str(words[0]), 'I’ve')
        self.assertEqual(str(words[1]), 'I’ve')

    def test_nesting(self):
        p = self.parse(["<p>", "I<b>", "’<i>ve</i> ", "I</b>’ve</p>"])
        words = p.words
        self.assertEqual(len(words), 2)
        self.assertEqual(str(words[0]), 'I’ve')
        self.assertEqual(str(words[1]), 'I’ve')

    def test_br_tag(self):
        p = self.parse("<p>Break tag<br/>should close word.</p>")
        words = p.words
        self.assertEqual(len(words), 6)
        self.assertEqual(str(words[1]), 'tag')
        self.assertEqual(words[2], Break)
        self.assertEqual(str(words[3]), 'should')

    def test_nbsp_ignored(self):
        p = self.parse("<p>Non &nbsp; breaking space.</p>")
        words = p.words
        self.assertEqual(len(words), 3)
        self.assertEqual(str(words[0]), 'Non')
        self.assertEqual(str(words[1]), 'breaking')


class TestRendering(BaseTestCase):

    def test_wrap(self):
        p = self.parse(
            "<p>Some people don't like change, but you need to "
            "embrace change if the alternative is disaster.</p>"
        )
        self.assertEqual(p.wrap(100), (100, 84))

    def test_draw_merges_same_styled_words(self):
        s = Style.default()
        self.assertPDF(Paragraph([
            Word(s, 'hello'),
            Word(s, 'world').add(s, '!')
        ], s), [
            "BT",
            "1 0 0 1 0 0 Tm",
            "/F1 12 Tf 14 TL",
            "0 0 Td",
            "(hello world!) Tj T*",
            "ET"
        ])

    def test_draw_optimally_switches_fonts(self):
        # merge continous fragments of the same style
        # into a single Tj draw command
        # <p>hello <b>world</b>, systori<b>!</b></p>
        s = Style.default()
        self.assertPDF(Paragraph([
            Word(s, 'hello'),
            Word(s.set(bold=True), 'world').add(s, ','),
            Word(s, 'systori').add(s.set(bold=True), '!'),
        ], s), [
            "BT",
            "1 0 0 1 0 0 Tm",
            "/F1 12 Tf 14 TL",
            "0 0 Td",
            "(hello ) Tj",
            "/F2 12 Tf",
            "(world) Tj",
            "/F1 12 Tf",
            "(, systori) Tj",
            "/F2 12 Tf",
            "(!) Tj T*",
            "ET"
        ])



import os.path
from utils import BaseTestCase
from reportlab import rl_config
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from bericht.pdf import PDFDocument

rl_config.TTFSearchPath = [os.path.join(os.path.dirname(__file__), 'fonts')]

font_name = "Ubuntu"
for font_face in ['Regular', 'Italic', 'Bold', 'BoldItalic']:
    font_family_member = '{}-{}'.format(font_name, font_face)
    pdfmetrics.registerFont(TTFont(font_family_member, font_family_member+'.ttf'))

pdfmetrics.registerFontFamily(
    font_name,
    normal='{}-Regular'.format(font_name),
    bold='{}-Bold'.format(font_name),
    italic='{}-Italic'.format(font_name),
    boldItalic='{}-BoldItalic'.format(font_name)
)


class TestFont(BaseTestCase):

    def render(self, html, css=None):
        parser = self.parse(html, css)
        p = next(iter(parser))
        doc = PDFDocument(parser.css)
        page = doc.add_page()
        p.wrap(page, page.available_width)
        p.draw(page, page.x, page.y)
        for _ in doc.header():
            pass
        for _ in page.read():
            pass
        for _ in doc.footer():
            pass

    def test_uninitialized_font_regression_helvetica(self):
        self.render("<table><tr><td>some text</td><td>&#931;</td></tr></table>")

    def test_uninitialized_font_regression_ubuntu(self):
        self.render("<table><tr><td>some text</td><td>&#931;</td></tr></table>",
                    "p { font-family: Ubuntu; }")



from bericht.pdf import PDFDocument
from utils import BaseTestCase


TABLE_HTML = """
<table>
  <colgroup>
    <col width="0*">
    <col width="1*">
    <col width="0*">
    <col width="0*">
  <thead>
    <tr>
      <td>code
      <td>name
      <td>price
      <td>total
  </thead>
  <tfoot>
    <tr><td colspan="4">footer<br>line
  </tfoot>
  <tr>
    <td>1
    <td>First Item
    <td>10.00
    <td>10.00
  <tr>
    <td>2
    <td>{}
    <td>15.00
    <td>25.00
""".format('Second Long Item '*1000)


class TestTable(BaseTestCase):

    def test_only_returns_rows(self):
        nodes = list(self.parse(TABLE_HTML))
        self.assertEqual(nodes[0].tag, 'tr')
        self.assertEqual(nodes[1].tag, 'tr')
        self.assertEqual(len(nodes), 2)

    def test_minimal_tags(self):
        nodes = self.parse(TABLE_HTML)
        row1, row2 = nodes
        page = PDFDocument(nodes.css).add_page()
        row1.wrap(page, page.available_width)
        row1.draw(page, page.x, page.y)


class TestCells(BaseTestCase):

    def get_cell(self, content):
        return next(iter(self.parse(
            """<table><tr><td>{}</td></tr></table>""".format(content)
        ))).children[0]

    def get_cells(self, content):
        return next(iter(self.parse(
            """<table><tr>{}</tr></table>""".format(''.join('<td>'+td+'<td>' for td in content))
        ))).children

    def test_blank_cell(self):
        cell = self.get_cell('')
        self.assertEqual(cell.children, [])

    def test_implicit_paragraph(self):
        cell = self.get_cell('hello world')
        self.assertEqual(cell.children, ['hello world'])

    def test_one_paragraph(self):
        cell = self.get_cell('<p>hello world')
        self.assertEqual(len(cell.children), 1)
        self.assertEqual(cell.children[0].children, ['hello world'])

    def test_two_paragraph(self):
        cell = self.get_cell('<p>hello<p>world')
        self.assertEqual(len(cell.children), 2)
        p1, p2 = cell.children
        self.assertEqual(p1.children, ['hello'])
        self.assertEqual(p2.children, ['world'])

    def test_br_cell(self):
        cell = self.get_cell('<p>hello<br>world')
        self.assertEqual(len(cell.children), 1)
        p = cell.children[0]
        self.assertEqual(len(p.children), 3)
        self.assertEqual(p.children[0], 'hello')
        self.assertEqual(p.children[1].tag, 'br')
        self.assertEqual(p.children[2], 'world')

    def test_recovery(self):
        cell = self.get_cell('<p><p>hello<p><b>world</p>!')
        self.assertEqual(len(cell.children), 4)
        p1, p2, p3, x = cell.children
        self.assertEqual(p1.children, [])
        self.assertEqual(p2.children, ['hello'])
        self.assertEqual(p3.children[0].tag, 'b')
        self.assertEqual(p3.children[0].children, ['world'])
        self.assertEqual(x, '!')


class TestRows(BaseTestCase):

    def get_rows(self, rows):
        return list(self.parse(
            """<table>{}</table>""".format(''.join('<tr><td>'+row+'<td></tr>' for row in rows))
        ))

    def test_two_rows(self):
        rows = self.get_rows(['hello', 'world'])
        self.assertEqual(len(rows), 2)



from unittest import TestCase, skip
from tinycss2 import parse_one_rule

from bericht.html.css import parse_selectors
from utils import BaseTestCase


class TestSelectorParsing(TestCase):

    def tokens(self, selector):
        return parse_one_rule(selector+'{}', skip_comments=True).prelude

    def parse(self, selector):
        return parse_selectors(self.tokens(selector))

    def assert_selector(self, tag, selectors=1, combinators=1, simple=1):
        sel = self.parse(tag)
        self.assertEqual(len(sel), selectors)
        self.assertEqual(len(sel[0].combinators), combinators)
        comb = sel[0].combinators[0]
        self.assertEqual(len(comb.simple_selectors), simple)

    def test_simple_tag(self):
        self.assert_selector(' p ')
        self.assert_selector('p ')
        self.assert_selector(' p')

    def test_multiple_selectors(self):
        self.assert_selector('p,b ', 2)
        self.assert_selector('p, b ', 2)
        self.assert_selector('p, b', 2)
        self.assert_selector('p,b', 2)

    def test_multiple_combinators(self):
        self.assert_selector('p b', 1, 2)
        self.assert_selector(' p  b ', 1, 2)
        self.assert_selector(' p >b ', 1, 2)
        self.assert_selector('p>b.foo', 1, 2)
        self.assert_selector('p>b>foo', 1, 3)
        self.assert_selector('p+b~foo', 1, 3)


class TestMatch(BaseTestCase):

    def test_class_tag_id_selectors(self):
        nodes = self.parse("""
            <p>initial</p>
            <p class="bolded">bolded</p>
            <p class="big bolded">big bolded</p>
            <p id="withid">small</p>
        """, """
            #withid, .bolded { font-size: 14pt; }
            .bolded { font-weight: bold; }
            p.big { font-size: 16pt; }
        """)
        initial, bolded, big, withid = nodes
        self.assertEqual(initial.style.font_size, 12)
        self.assertEqual(initial.style.font_weight, '')
        self.assertEqual(bolded.style.font_size, 14)
        self.assertEqual(bolded.style.font_weight, 'bold')
        self.assertEqual(big.style.font_size, 16)
        self.assertEqual(big.style.font_weight, 'bold')
        self.assertEqual(withid.style.font_size, 14)
        self.assertEqual(withid.style.font_weight, '')

    def test_combinators(self):
        nodes = list(self.parse("""
            <p><b>text</b></p>
            <div><b>text</b></div>
            <div><p><b>text</b></p></div>
        """, """
            b { font-size: 14pt; }
            div b { font-size: 16pt; }
            div > b { color: green; }
        """))
        p_b = nodes[0].children[0]
        div_b = nodes[1].children[0]
        div_p_b = nodes[2].children[0].children[0]
        self.assertEqual(p_b.style.font_size, 14)
        self.assertEqual(div_b.style.font_size, 16)
        #self.assertEqual(div_p_b.style.font_size, 16)

    def test_odd_even_row_background(self):
        nodes = list(self.parse("""
        <table>
          <tbody>
            <tr><td>odd row</td></tr>
            <tr><td>even row</td></tr>
            <tr><td>odd 2 row</td></tr>
            <tr><td>even 2 row</td></tr>
          </tbody> 
        </table>
        """, """
        table > tbody > tr:nth-child(even) {
          background-color: green;
        }
        """))
        odd, even, odd2, even2 = nodes
        self.assertIsNone(odd.style.background_color)
        self.assertIsNotNone(even.style.background_color)
        self.assertIsNone(odd2.style.background_color)
        self.assertIsNotNone(even2.style.background_color)
